/* Hack'n'Slash, A generic MUD engine.
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL tribe
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <hacknslash/array.h>

typedef struct ArrayElem ArrayElem_t;
struct ArrayElem {
  gpointer thing;
  gint     id;
};

static
ArrayElem_t * _array_elem_new ( Array_t * array, gpointer thing, gint id ) {
  ArrayElem_t * elem = g_new( ArrayElem_t, 1 );
  elem->thing = thing;
  elem->id = id;
  if( array->on_add )
    array->on_add( thing );
  return elem;
}

static
void          _array_elem_free( Array_t * array, ArrayElem_t * elem ) {
  if( array->on_remove )
    array->on_remove( elem->thing );
  g_free( elem );
}



/* */
void         array_init ( Array_t *    array,
			  guint        size,
			  Func_t       on_add,
			  Func_t       on_remove ) {
  array->size = size;
  array->array = g_new0( GSList *, size );
  array->on_add = on_add;
  array->on_remove = on_remove;
}

Array_t *    array_new  ( guint        size,
			  Func_t       on_add,
			  Func_t       on_remove ) {
  Array_t * array = g_new( Array_t, 1 );
  array_init( array, size, on_add, on_remove );
  return array;
}

void         array_free ( Array_t * array ) {
  if( !array ) return;
  g_free( array->array );
  g_free( array );
}

void         array_clear( Array_t *     array ) {
  gint i;
  GSList * slist;
  for( i = 0; i < array->size; i++ ) {
    for( slist = array->array[i]; slist; slist = slist->next )
      _array_elem_free( array, slist->data );
    g_slist_free( array->array[i] );
    array->array[i] = NULL;
  }
}

void      array_store( Array_t * array,
		       gpointer  thing,
		       gint      id ) {
  gint index;
  if( !array || !(array)->size || !thing ) return;
  array_remove( array, id );
  index = id % array->size;
  array->array[index] = g_slist_prepend( array->array[index],
					 _array_elem_new( array, thing, id ) );
}

static
GSList * _array_get  ( Array_t * array,
		       gint      id,
		       GSList ** trail ) {
  gint index = id % array->size;
  GSList * slist;
  for( slist = array->array[index], *trail = NULL;
       slist;
       *trail = slist, slist = slist->next ) {
    ArrayElem_t * elem = slist->data;
    if( elem->id == id ) {
      return slist;
    }
  }
  return NULL;
}

gpointer  array_get   ( Array_t * array,
			gint      id ) {
  GSList * slist, * trail;
  slist = _array_get( array, id, &trail );
  if( slist )
    return ((ArrayElem_t *)(slist->data))->thing;
  return NULL;
}

void      array_remove( Array_t * array,
			gint      id ) {
  GSList * slist, * trail;
  gint index;
  if( !array ) return;
  index = id % array->size;
  slist = _array_get( array, id, &trail );
  if( slist ) {
    if( trail )
      trail->next = slist->next;
    else
      array->array[index] = slist->next;
    _array_elem_free( array, slist->data );
    g_slist_free_1( slist );
  }
}

void      array_foreach( Array_t *      array,
			 ArrayForFunc_t func,
			 gpointer       data ) {
  gint i;
  GSList * slist;
  if( !func || !array ) return;
  for( i = 0; i < array->size; i++ ) {
    for( slist = array->array[i]; slist; slist = slist->next ) {
      ArrayElem_t * elem = slist->data;
      func( elem->thing, elem->id, data );
    }
  }
}

gpointer  array_find   ( Array_t *       array,
			 ArrayFindFunc_t func,
			 gpointer        data ) {
  gint i;
  GSList * slist;
  if( !func || !array ) return NULL;
  for( i = 0; i < array->size; i++ ) {
    for( slist = array->array[i]; slist; slist = slist->next ) {
      ArrayElem_t * elem = slist->data;
      if( func( elem->thing, elem->id, data ) )
	return elem->thing;
    }
  }
  return NULL;
}
