/* Hack'n'Slash, A generic MUD engine.
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL tribe
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <unistd.h>

#include <hacknslash/misc.h>
#include <hacknslash/token.h>
#include <hacknslash/block.h>

#define BLOCK_SKIP " \t\b\r\n" /* stuff inbetween tokens */
#define BLOCK_QUOTE "~~\"\"`'" /* quote pairs */

/* */
BlockVar_t * block_var_new ( gchar * label, gchar * value ) {
  BlockVar_t * var;
  if( !label ) return NULL;
  var = g_new( BlockVar_t, 1 );
  var->label = g_strdup( label );
  var->value = g_strdup( value );
  return var;
}

void         block_var_free( BlockVar_t * var ) {
  if( !var ) return;
  g_free( var->label );
  g_free( var->value );
  g_free( var );
}



/* */
Block_t *    block_new ( gchar * label, gchar * id ) {
  Block_t * b;
  if( !label ) return NULL;
  b = g_new( Block_t, 1 );
  b->label = g_strdup( label );
  b->id = g_strdup( id );
  b->vars = b->vars_last = b->blocks = b->blocks_last = NULL;
  return b;
}

void         block_clear   ( Block_t * b ) {
  GSList * scan;
  for( scan = b->blocks; scan; scan = scan->next ) {
    if( scan->data )
      block_free( scan->data );
  }
  g_slist_free( b->blocks );
  b->blocks = b->blocks_last = NULL;
  for( scan = b->vars; scan; scan = scan->next ) {
    if( scan->data )
      block_var_free( scan->data );
  }
  g_slist_free( b->vars );
  b->vars = b->vars_last = NULL;
}

void         block_free    ( Block_t * b ) {
  if( b ) return;
  g_free( b->label );
  g_free( b->id );
  block_clear( b );
  g_free( b );
}

/* */
void         block_add_var  ( Block_t *    b,
			      BlockVar_t * var ) {
  GSList * new;
  new = g_slist_alloc();
  new->next = NULL;
  new->data = var;
  if( b->vars_last )
    b->vars_last->next = new;
  else
    b->vars = new;
  b->vars_last = new;
}

BlockVar_t * block_get_var  ( Block_t *    block,
			      gchar *      label ) {
  GSList * scan;
  if( !block || !label ) return NULL;
  for( scan = block->vars; scan; scan = scan->next ) {
    BlockVar_t * var = scan->data;
    if( !strcmp( var->label, label ) )
      return var;
  }
  return NULL;
}

/* */
void         block_add_block( Block_t *    b,
			      Block_t *    block ) {
  GSList * new;
  new = g_slist_alloc();
  new->next = NULL;
  new->data = block;
  if( b->blocks_last )
    b->blocks_last->next = new;
  else
    b->blocks = new;
  b->blocks_last = new;
}

Block_t *    block_get_block( Block_t *    block,
			      gchar *      label,
			      gchar *      id ) {
  GSList * scan;
  if( !block || !label ) return NULL;
  for( scan = block->blocks; scan; scan = scan->next ) {
    Block_t * b = scan->data;
    if( !id || !misc_strcmp( b->id, id ) )
      return b;
  }
  return NULL;
}

/* */
static
void _skip( gchar * skip, FILE * f ) {
  gint ch;
  gint l = 0;
  while( 1 ) {
    ch = fgetc( f );
    if( feof( f ) )
      break;
    l = 1;
    if( !strchr( skip, ch ) )
      break;
  }
  if( l )
    fseek( f, -1L, SEEK_CUR );
}

static
gint     _block_skip        ( FILE * f ) {
  gchar buf[3];
  gint len;
  gint count = 1;
  while( count > 0 ) {
    len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, buf, 3 );
    if( len < 0 )
      break;
    if( buf[0] == '{' && !(buf[1]) )
      count++;
    else if( buf[0] == '}' && !(buf[1]) )
      count--;
  }
  if( len >= 0 )
    _skip( BLOCK_SKIP, f );
  return count;
}

static
gboolean _contains( gchar * str, gchar * test ) {
  while( *str ) {
    if( strchr( test, *str ) )
      return TRUE;
    str++;
  }
  return FALSE;
}



/* */
static
Block_t * _block_file_getf( gchar * _label,
			    gchar * id,
			    FILE *  f ) {
  gint len;
  gchar label[32];
  gchar op[32];
  gchar tmp[3];
  gchar value[512], * _id;
  gint ch;
  Block_t * b = block_new( _label, id );

  while( 1 ) {
    len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, label, 32 );
    if( len < 0 )
      break;
    if( label[0] == '#' ) {
      /* skip comment */
      while( 1 ) {
        ch = fgetc( f );
        if( feof( f ) || ch == '\n' ) break;
      }
      continue;
    }

    if( label[0] == '}' && !label[1] )
      break;

    if( label[0] == '{' && !label[1] ) {
      /* syntax error */
      _block_skip( f );
      continue;
    }

    /* get the operator */
    len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, op, 32 );
    if( len <= 0 ) {
      break;
    }

    if( label[0] == '=' && !label[1] ) {
      /* syntax error, somehow found ourselves in the wrong place in an
       * assignment, so skip current one and move onto next.
       */
      continue;
    }

    if( op[0] == '=' && !op[1] ) {
      len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, value, 512 );
      if( len <= 0 )
        break;

      /* an assignment, add to list of vars */
      block_add_var( b, block_var_new( label, value ) );
      continue;

    } else {
      if( op[0] == '{' && !op[1] ) {
        _id = NULL;

      } else {
        len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, tmp, 3 );
        if( len <= 0 )
          break;
        if( tmp[0] != '{' || tmp[1] )
          continue;
        _id = op;
      }

      block_add_block( b, _block_file_getf( label, _id, f ) );
      continue;
    }
  }

  return b;
}

/* */
Block_t * block_file_getf( gchar * _label,
			   gchar * id,
			   FILE *  f ) {
  gint len;
  gchar label[32];
  gchar op[32];
  gchar tok[32];
  gchar * _id;
  gint ch;

  while( 1 ) {
    len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, label, 32 );
    if( len < 0 )
      break;

    if( label[0] == '}' && !(label[1]) )
      continue;

    if( label[0] == '#' ) {
      /* skip comment */
      while( 1 ) {
        ch = fgetc( f );
        if( feof( f ) ) 
	  return NULL;
	if( ch == '\n' ) 
	  break;
      }
      continue;
    }
    if( !label[1] && ( label[0] == '{' || label[0] == '=' ) ) {
      /* syntax error */
      _block_skip( f );
      continue;
    }

    len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, op, 32 );
    if( len <= 0 )
      break;

    if( !op[1] && ( op[0] == '}' || op[0] == '=' ) ) {
      _block_skip( f );
      continue;
    }

    if( !op[1] && op[0] == '{' ) {
      _id = NULL;
    } else {
      _id = op;
      len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, tok, 32 );
      if( len <= 0 )
        break;
      if( tok[1] || tok[0] != '{' )
        continue;
    }

    if( !strcmp( label, _label ) ) {
      if( !id || (id && _id && !strcmp( id, _id )))
        return _block_file_getf( label, _id, f );
    }

    /* skip current block as does not match label and or id */
    _block_skip( f );
  }

  return NULL;
}

Block_t * block_file_get ( gchar *      label,
			   gchar *      id,
			   gchar *      file ) {
  FILE * f;
  Block_t * b = NULL;
  f = fopen( file, "r" );
  if( f ) {
    b = block_file_getf( label, id, f );
    fclose( f );
  }
  return b;
}

static
void    token_fputs( gchar * str, FILE * f ) {
  if( !str )
    fputs( "~~", f );
  else if(_contains( str, BLOCK_SKIP ) ) {
    fputc( '~', f );
    fputs( str, f );
    fputc( '~', f );
  } else
    fputs( str, f );
}

static
void        _block_file_write    ( Block_t * block,
				   gint      indent,
				   FILE *    f ) {
  GSList * scan;

  if( block->label == NULL ) return;

  /* write block head */
  misc_findent( indent, f );
  token_fputs( block->label, f );
  if( block->id ) {
    fputc( ' ', f );
    token_fputs( block->id, f );
  }
  fputs( " {\n", f );

  /* write vars */
  indent+=2;
  for( scan = block->vars; scan; scan = scan->next ) {
    BlockVar_t * var = scan->data;
    if( var->label ) {
      misc_findent( indent, f );
      token_fputs( var->label, f );
      fputs( " = ", f );
      token_fputs( var->value, f );
      fputc( '\n', f );
    }
  }

  /* write subblocks */
  for( scan = block->blocks; scan; scan = scan->next )
    _block_file_write( scan->data, indent, f );

  /* write tail */
  misc_findent( indent-2, f );
  fputs( "}\n", f );
}

void         block_file_write    ( Block_t * block,
				   FILE *    f ) {
  _block_file_write( block, 0, f );
}

static
void        _block_file_insert   ( Block_t * block,
				   gchar *  _label,
				   gchar *   id,
				   gchar *   file ) {
  gint len;
  gchar label[32];
  gchar op[32];
  gchar tok[32];
  gchar * _id;
  gint open_blocks = 0, ch;
  long off;
  FILE * f, * tmp;
  gchar * tmpfile = ".tmp";
  gint done = ( block ? 0 : 1 );

  f = fopen( file, "r+" );
  if( !f ) {
    f = fopen( file, "w" );
    if( !f ) {
      perror( "fopen" );
      return;
    }
    _block_file_write( block, 0, f );
    fclose( f );
    return;
  }

  while( 1 ) {
    _skip( BLOCK_SKIP, f );
    off = ftell( f );

    len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, label, 32 );
    if( len <= 0 )
      break;

    if( label[0] == '}' && !label[1] )
      continue;

    if( label[0] == '#' ) {
      /* skip comment */
      while( 1 ) {
        ch = fgetc( f );
        if( feof( f ) ) break;
        if( ch == '\n' ) break;
      }
      continue;
    }

    if( !label[1] && ( label[0] == '{' || label[0] == '=' ) ) {
      /* syntax error */
      open_blocks = _block_skip( f );
      continue;
    }

    len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, op, 32 );
    if( len <= 0 )
      break;

    if( !op[1] && op[0] == '}' ) {
      open_blocks = _block_skip( f );
      continue;
    }

    if( !op[1] && op[0] == '=' ) {
      open_blocks = _block_skip( f );
      continue;
    }

    if( !op[1] && op[0] == '{' ) {
      _id = NULL;

    } else {
      _id = op;
      len = token_copy_f( f, BLOCK_SKIP, BLOCK_QUOTE, tok, 32 );
      if( len <= 0 )
        break;
      if( tok[1] || tok[0] != '{' )
        continue;
    }

    if( !strcmp( label, _label ) ) {
      if( (!id && !_id) ||
          (id && _id && !strcmp( id, _id ) ) ) {
        /* open tmp file */
        tmp = fopen( tmpfile, "w+" );
        /* skip passed old block */
        _block_skip( f );
        /* save rest of file to tmp */
        misc_fcopy( f, tmp, -1 );
        /* go back to beginning of old block */
        fseek( f, off, SEEK_SET );
        /* overwrite old block */
        if( !done ) {
          _block_file_write( block, 0, f );
          done = 1;
        }
        /* rewrite rest of file */
        rewind( tmp );
        misc_fcopy( tmp, f, -1 );
	fclose( tmp );
	unlink( tmpfile );
        break;
      }
    }
    /* skip current block as does not match label and or id */
    open_blocks = _block_skip( f );
  }

  while( open_blocks ) {
    open_blocks--;
    misc_findent( open_blocks*2, f );
    fputs( "}\n", f );
  }

  if( !done ) {
    fputc( '\n', f );
    _block_file_write( block, 0, f );
  }

  off = ftell( f );
  fclose( f );
  truncate( file, off );
}

void         block_file_insert       ( Block_t * block,
				       gchar *   file ) {
  _block_file_insert( block, block->label, block->id, file );
}

void         block_file_remove       ( gchar *   label,
				       gchar *   id,
				       gchar *   file ) {
  _block_file_insert( NULL, label, id, file );
}

/* */
Object_t *   block_to_object         ( Block_t *    block,
				       BlockDef_t * def ) {
  Object_t * obj;
  GSList * scan1, * scan2;
  ObjectClass_t * oclass = def->oclass;

  obj = object_new( oclass, block->id );
  
  if( def->add_var ) {
    for( scan1 = block->vars; scan1; scan1 = scan1->next ) {
      BlockVar_t * var = scan1->data;
      if( FALSE == def->add_var( obj, var ) )
	; /* do something with unrecognised var? */
    }
  }

  for( scan1 = block->blocks; scan1; scan1 = scan1->next ) {
    Block_t * tmp_block = scan1->data;
    for( scan2 = def->blocks; scan2; scan2 = scan2->next ) {
      BlockDef_t * tmp_def = scan2->data;
      if( !strcmp( tmp_block->label, tmp_def->oclass->name ) ) {
	object_add( obj, block_to_object( tmp_block, tmp_def ) );
	break;
      }
    }
  }

  return obj;
}

Block_t *    block_from_object       ( Object_t *   object,
				       BlockDef_t * def ) {
  Block_t * block;
  GSList * scan1, * scan2;
  ObjectClass_t * oclass = (ObjectClass_t *) (((Object_t *)block)->oclass);

  block = block_new( oclass->name, object->id );
  if( def->get_vars )
    block->vars = def->get_vars( object );  

  for( scan1 = object->objects; scan1; scan1 = scan1->next ) {
    Object_t * o = scan1->data;
    for( scan2 = def->blocks; scan2; scan2 = scan2->next ) {
      BlockDef_t * tmp_def = scan2->data;
      if( o->oclass == tmp_def->oclass ) {
	block_add_block( block, block_from_object( o, tmp_def ) );
	break;
      }
    }
  }

  return block;
}
