/* Hack'n'Slash, A generic MUD engine.
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL tribe
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __OBJECT_H__
#define __OBJECT_H__

#include <glib.h>

typedef struct ObjectClass ObjectClass_t;
typedef struct Object      Object_t;

typedef void     (* ObjectFunc_t)      ( Object_t * obj );
typedef void     (* ObjectDataFunc_t)  ( Object_t * obj, gpointer data );
typedef gboolean (* ObjectFindFunc_t)  ( Object_t * obj, gpointer data );

#include <hacknslash/defs.h>
#include <hacknslash/block.h>



/* */
struct ObjectClass {
  /* the name of this class */
  gchar * name;

  /* the parent class of this class */
  ObjectClass_t * parent_class;

  /* size of each object */
  guint size;

  /* init() is called when the object is first created */
  ObjectFunc_t init;

  /* destroy is called when the object is destroyed */
  ObjectFunc_t destroy;
};

/* returns the class of type Object_t */
ObjectClass_t * object_class    ();



/* */
struct Object {
  ObjectClass_t * oclass;
  guint ref_count;
  gchar * id;

  GSList * objects, * objects_last; /* child objects */

  GHookList hooks; /* hook list (for hooking object events) */
};



/* create a new object of class `oclass' with id `id' */
Object_t *      object_new        ( ObjectClass_t *   oclass,
				    gchar *           id );

/* reference/unreferece `obj' */
void            object_ref        ( Object_t *        obj );
void            object_unref      ( Object_t *        obj );

/* retrieve the first found object from `obj' with class `oclass' and 
 * id `id' or NULL if not found.
 */
Object_t *      object_get        ( Object_t *        obj,
				    ObjectClass_t *   oclass,
				    gchar *           id );

/* remove the first object from `obj' found with class `oclass' and id `id' */
void            object_remove     ( Object_t *        obj,
				    ObjectClass_t *   oclass,
				    gchar *           id );

/* add `add' to `obj' */
void            object_add        ( Object_t *        obj,
				    Object_t *        add );

/* calls `func' on each child object of `obj' with class `oclass' (if
 * `oclass' is not NULL)
 */
void            object_foreach    ( Object_t *        obj,
				    ObjectClass_t *   oclass,
				    ObjectDataFunc_t  func,
				    gpointer          data );

/* use `func' to find an object of type `oclass' (or all object types
 * if `oclass' is NULL) and return it.
 */
Object_t *      object_find       ( Object_t *        obj,
				    ObjectClass_t *   oclass,
				    ObjectFindFunc_t  func,
				    gpointer          data );



/* return object defined in `def' with `id' from source in `file' 
 * calls `block_file_get()' and `block_to_object()'
 */
Object_t *      object_file_get   ( BlockDef_t *      def,
				    gchar *           id,
				    gchar *           file );

/* insert `obj' into `file',
 * calls `block_from_object()' and `block_file_insert()' 
 */
void            object_file_insert( Object_t *        obj,
				    BlockDef_t *      def,
				    gchar *           file );
				    
#define         object_file_remove( _OCLASS_, _ID_, _FILE_ ) \
  block_file_remove( (_OCLASS_)->name, (_ID_), (_FILE_) )



/* */
enum {
  OBJECT_EVENT_DESTROY = 0,

  NUM_OBJECT_EVENTS
};

/* returns real event type from enumerated type */
guint           object_event_type( guint type );

typedef struct ObjectEvent ObjectEvent_t;
struct ObjectEvent {
  guint type;
  /* derive from `ObjectEvent' to add more stuff */
};

typedef gboolean (* ObjectEventFunc_t)( Object_t *      obj,
					ObjectEvent_t * event,
					gpointer        data );

#define object_hook( _OBJ, _SIG, _FUNC, _DATA ) \
  object_hook_full( (_OBJ), (_SIG), (_FUNC), (_DATA), NULL )

guint           object_hook_full( Object_t *        obj,
				  guint             type,
				  ObjectEventFunc_t func,
				  gpointer          data,
				  Func_t            destroy );

void            object_unhook   ( Object_t *        obj,
				  guint             hook );

void            object_emit     ( Object_t *        obj,
				  ObjectEvent_t *   event );

/* allocate a consecutive block of `num' unique event type */
void            object_event_type_alloc( guint * array,
					 guint   num );

#endif /* __OBJECT_H__ */









/* documentation is for the weak */
