/* Hack'n'Slash, A generic MUD engine.
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL tribe
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* object.c: 
 *   this object structure was inspired by GtkObject
 */

#include <hacknslash/object.h>
#include <hacknslash/misc.h>

#include <stdio.h>
#include <stdlib.h>

typedef struct _ObjectHook _ObjectHook_t;
struct _ObjectHook {
  GHook parent;
  guint type; /* event type */
};

static
guint _object_events[ NUM_OBJECT_EVENTS ];

guint           object_event_type( guint type ) {
  static gboolean ran = FALSE;
  if( !ran ) {
    ran = TRUE;
    object_event_type_alloc( _object_events, NUM_OBJECT_EVENTS );
  }
  return ( type >= 0 && type < NUM_OBJECT_EVENTS ?
	   _object_events[ type ] :
	   0 );
}

/* */
static
void           _object_init   ( Object_t * obj ) {
  obj->ref_count = 0;
  obj->id = NULL;
  obj->objects = obj->objects_last = NULL;
  g_hook_list_init( &(obj->hooks), sizeof( _ObjectHook_t ) );
}

static
void           _object_destroy( Object_t * obj ) {
  GSList * slist;
  g_hook_list_clear( &(obj->hooks) );
  for( slist = obj->objects ; slist; slist = slist->next )
    object_unref( slist->data );
  g_slist_free( obj->objects );
  g_free( obj->id );
}

static ObjectClass_t _oclass = {
  "Object",
  NULL,
  sizeof( Object_t ),
  _object_init,
  _object_destroy
};

ObjectClass_t * object_class() {
  return &_oclass;
}

/* */
Object_t *      object_new  ( ObjectClass_t * oclass,
			      gchar *         id ) {
  Object_t * obj;
  if( !oclass )
    oclass = object_class();
  obj = g_malloc0( MAX( sizeof( Object_t ), oclass->size ) );
  obj->oclass = oclass;
  if( oclass->init )
    oclass->init( obj );
  obj->id = g_strdup( id );
  return obj;
}

void            object_free ( Object_t *      obj ) {
  /* emit destroy event */
  ObjectEvent_t event = { object_event_type( OBJECT_EVENT_DESTROY ) };
  object_emit( obj, &event );

  /* actually destroy the object */
  if( obj->oclass->destroy )
    obj->oclass->destroy( obj );
  g_free( obj );
}

void            object_ref  ( Object_t *      obj ) {
  if( !obj ) return;
  obj->ref_count++;
}

void            object_unref( Object_t *      obj ) {
  if( !obj ) return;
  if( obj->ref_count > 1 )
    obj->ref_count --;
  else {
    if( obj->ref_count == 0 )
      g_print( "Hack'n'Slash error: \"%s\" object unref'd more times than ref'd.\n",
	       obj->oclass->name );
    else
      object_free( obj );
  }
}

/* */

static
GSList *       _object_find    ( Object_t *      obj,
				 ObjectClass_t * oclass,
				 gchar *         id,
				 GSList **       trail ) {
  GSList * scan;
  *trail = NULL;
  for( scan = obj->objects; scan; *trail = scan, scan = scan->next ) {
    Object_t * o = scan->data;
    if( (!oclass || (o->oclass == oclass)) && !misc_strcmp( o->id, id ))
      return scan;
  }
  return NULL;
}

Object_t *      object_get     ( Object_t *      obj,
				 ObjectClass_t * oclass,
				 gchar *         id ) {
  GSList * slist, * trail;
  if( NULL != (slist = _object_find( obj, oclass, id, &trail )))
    return slist->data;
  return NULL;
}

void            object_foreach ( Object_t *       obj,
				 ObjectClass_t *  oclass,
				 ObjectDataFunc_t func,
				 gpointer         data ) {
  GSList * scan;
  for( scan = obj->objects; scan; scan = scan->next ) {
    Object_t * o = scan->data;
    if( !oclass || o->oclass == oclass )
      func( o, data );
  }
}

Object_t *      object_find    ( Object_t *       obj,
				 ObjectClass_t *  oclass,
				 ObjectFindFunc_t func,
				 gpointer         data ) {
  GSList * scan;
  for( scan = obj->objects; scan; scan = scan->next ) {
    Object_t * o = scan->data;
    if( !oclass || o->oclass == oclass ) 
      if( func( o, data ) )
	return o;
  }
  return NULL;
}

void            object_remove  ( Object_t *      obj,
				 ObjectClass_t * oclass,
				 gchar *         id ) {
  GSList * slist, * trail;
  if( NULL != (slist = _object_find( obj, oclass, id, &trail ))) {
    if( trail )
      trail->next = slist->next;
    else
      obj->objects = slist->next;
    object_unref( slist->data );
    g_slist_free_1( slist );
  }
}

void            object_add     ( Object_t *      obj,
				 Object_t *      add ) {
  GSList * new;
  new = misc_slist_new( add );
  if( obj->objects_last )
    obj->objects_last->next = new;
  else
    obj->objects = new;
  obj->objects_last = new;
  object_ref( add );
}

Object_t *      object_file_get( BlockDef_t *     def,
				 gchar *          id,
				 gchar *          file ) {
  Block_t * b;
  if( NULL != (b = block_file_get( def->oclass->name, id, file ))) {
    Object_t * o;
    o = block_to_object( b, def );
    block_free( b );
    return o;
  }
  return NULL;
}

void            object_file_insert( Object_t *        obj,
				    BlockDef_t *      def,
				    gchar *           file ) {
  Block_t * b;
  b = block_from_object( def, obj );
  block_file_insert( b, file );
  block_free( b );
}



/* object event hooks */
guint           object_hook_full( Object_t *        obj,
				  guint             type,
				  ObjectEventFunc_t func,
				  gpointer          data,
				  Func_t            destroy ) {
  _ObjectHook_t * hook;
  GHook * ghook;
  hook = (_ObjectHook_t *) (ghook = g_hook_alloc( &(obj->hooks) ));
  hook->type = type;
  ghook->func = func;
  ghook->data = data;
  ghook->destroy = destroy;
  g_hook_prepend( &(obj->hooks), ghook );
  return ghook->hook_id;
}

void            object_unhook   ( Object_t *        obj,
				  guint             hook ) {
  g_hook_destroy( &(obj->hooks), hook );
}

/* */
typedef struct _ObjectInvoke _ObjectInvoke_t;
struct _ObjectInvoke {
  Object_t * obj;
  ObjectEvent_t * event;
};

static
gboolean _object_hook_marshaller( GHook * ghook, _ObjectInvoke_t * info ) {
  _ObjectHook_t * hook = (_ObjectHook_t *)ghook;
  if( info->event->type == hook->type )
    return ((ObjectEventFunc_t)(ghook->func))( info->obj, info->event, ghook->data );
  return TRUE;
}

void            object_emit     ( Object_t *        obj,
				  ObjectEvent_t *   event ) {
  _ObjectInvoke_t info = { obj, event };
  g_hook_list_marshal_check( &(obj->hooks),
			     TRUE,
			     (GHookCheckMarshaller) _object_hook_marshaller,
			     &info );
}

void            object_event_type_alloc( guint * array,
					 guint   num ) {
  static guint _next_id = 1;
  while( num ) {
    /* assign `_next_id' to *array, then increment `_next_id' and `array' */
    *(array++) = _next_id++;
    num--;
  }
}
