/* Hack'n'Slash, A generic MUD engine.
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL tribe
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* server.c: a dumb, barely functional socket server */

#include <arpa/telnet.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>

#include <hacknslash/server.h>
#include <hacknslash/misc.h>
#include <hacknslash/io.h>

/* why? no reason */
#define DEFAULT_PORT 4950

extern gint errno;

static gint _listen_fd = -1;
static gint _listen_port = DEFAULT_PORT;
static gint _listen_handler = 0;

/* database of all connected sockets */
static Sock_t * _socks[ FD_SETSIZE ];
static gint _num_clients = 0;
static gint _max_clients = 0;

/* handlers */
static SockFunc_t _connect_func = NULL;
static gpointer   _connect_data = NULL;
static SockFunc_t _disconnect_func = NULL;
static gpointer   _disconnect_data = NULL;
static SockFunc_t _read_func = NULL;
static gpointer   _read_data = NULL;

/* class of object to use as sockets */
static SockClass_t * _sock_class = NULL;



/* base class for object `Sock' */
static SockClass_t     _base_sock_class;

/* `Sock' is derived from `Object' */
static ObjectClass_t * _parent_class = NULL;

/* */
static
void _sock_init( Sock_t * sock ) {
  if( _parent_class->init )
    _parent_class->init( (Object_t *)sock );
  sock->fd = 0;
  sock->read_handle = 0;
  sock->line[0] = 0;
  sock->line_i = 0;
  sock->colors = FALSE;
  sock->putchar = NULL;
}

static
void _sock_destroy( Sock_t * sock ) {
  if( sock->fd >= 0 )
    close( sock->fd );
  if( sock->read_handle )
    io_unhook( sock->read_handle );
  if( _parent_class->destroy )
    _parent_class->destroy( (Object_t *)sock );
}

SockClass_t * sock_class() {
  if( !_parent_class ) {
    ObjectClass_t * oclass;
    _parent_class = object_class();
    oclass = (ObjectClass_t *) &_base_sock_class;
    oclass->name = "Sock";
    oclass->parent_class = _parent_class;
    oclass->size = sizeof( Sock_t );
    oclass->init = (ObjectFunc_t)_sock_init;
    oclass->destroy = (ObjectFunc_t)_sock_destroy;
  }
  return &_base_sock_class;
}



/* initialize the server */
void server_init          ( SockClass_t * sclass ) {
  static gint run = 0;
  if( run ) return;
  run = 1;
  _sock_class = ( sclass ? sclass : sock_class() );
}



/* input handling on a per character basis */
static
void _get_do      ( Sock_t * sock, guchar ch ) {
  switch( ch ) {
  case TELOPT_ECHO:
    /* do something ? */
    break;

  default:
    /* do something ? */
  }
  sock->putchar = NULL;
}

static
void _get_dont    ( Sock_t * sock, guchar ch ) {
  switch( ch ) {
  case TELOPT_ECHO:
    /* do something ? */
    break;

  default:
    /* do something ? */
  }
  sock->putchar = NULL;
}

static
void _get_iac     ( Sock_t * sock, guchar ch ) {
  switch( ch ) {
  case DO:
    sock->putchar = _get_do;
    break;

  case DONT:
    sock->putchar = _get_dont;
    break;

  default:
    /* broken! */
    sock->putchar = NULL;
  }
}

/* normal char handler */
static
void _get_char ( Sock_t * sock, guchar ch ) {
  if( sock->line_i > SOCK_LINE_SIZE ) {
    /* throw away until newline is found */
    if( ch == '\n' ) {
      sock->line[ 0 ] = 0;
      sock->line_i = 0;
    }
  } else {
    switch( ch ) {
    case IAC:
      /* read telnet command */
      sock->putchar = _get_iac;
      break;

    case '\033':
    case '\r':
      break;

    default:
      if( ch == '\n' || sock->line_i == SOCK_LINE_SIZE )
	ch = 0;
      sock->line[ sock->line_i ] = ch;
      sock->line_i++;
      if( !ch ) {
	if( _read_func )
	  _read_func( sock, _read_data );
	if( sock->line_i < SOCK_LINE_SIZE ) {
	  sock->line[ 0 ] = 0;
	  sock->line_i = 0;
	}
      }
    }
  }
}

static
void _server_read         ( gint fd, gint cond, gpointer data ) {
  guchar ch;
  Sock_t * sock = _socks[fd];
  int len;

  /* check to see if socket has disconnected */
  if( ioctl( fd, FIONREAD, &len ) < 0 || !len ) {
    server_disconnect( sock );
    return;
  }

  while( len > 0 ) {
    if( 0 >= read( fd, &ch, 1 ) ) {
      server_disconnect( sock );
      return;
    }

    if( sock->putchar )
      sock->putchar( sock, ch );
    else
      _get_char( sock, ch );
    len--;
  }
}

static
void _server_accept       ( gint fd, gint cond, gpointer data ) {
  gint newfd;
  struct sockaddr_in addr;
  gint addr_len = sizeof( addr );
  Sock_t * sock;
  gchar id[16];

  if( 0 > (newfd = accept( fd, (struct sockaddr *) &addr, &addr_len )) ) {
    perror( "accept" );
    return;
  }
  if( _max_clients > 0 && _num_clients >= _max_clients ) {
    close( fd );
    return;
  }
  if( 0 > fcntl( newfd, F_SETFL, O_NONBLOCK ) ) {
    perror( "fcntl" );
    exit( EXIT_FAILURE );
  }

  snprintf( id, 16, "%d", newfd );
  sock = (Sock_t *)object_new( (ObjectClass_t *)_sock_class, id );
  sock->fd = newfd;
  sock->read_handle = io_hook( newfd, IO_READ, _server_read, NULL, NULL );
  object_ref( (Object_t *)sock );
  _socks[ newfd ] = sock;
  _num_clients++;

  /* call the connect handler */
  if( _connect_func )
    _connect_func( sock, _connect_data );
}

/* open the server listen socket on port `port' */
void server_open         ( gint       port,
			   gint       _max_clients ) {
  struct sockaddr_in name;
  gint val;

  server_close();
  if( port > 0 )
    _listen_port = port;

  if( 0 > (_listen_fd = socket( PF_INET, SOCK_STREAM, 0 )) ) {
    /* error! */
    perror( "socket" );
    exit( EXIT_FAILURE );
  }

  if( 0 > fcntl( _listen_fd, F_SETFL, O_NONBLOCK ) ) {
    perror( "fcntl" );
    exit( EXIT_FAILURE );
  }

  val = 1;
  if( 0 > setsockopt( _listen_fd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof( val ) ) ) {
    perror( "setsockopt" );
    exit( EXIT_FAILURE );
  }

  name.sin_family = AF_INET;
  name.sin_port = htons( _listen_port );
  name.sin_addr.s_addr = htonl( INADDR_ANY );
  if( 0 > bind( _listen_fd, (struct sockaddr *) &name, sizeof(name) )) {
    perror( "bind" );
    exit( EXIT_FAILURE );
  }

  if( 0 > listen( _listen_fd, 5 ) ) {
    perror( "listen" );
    exit( EXIT_FAILURE );
  }

  /* listen for connect events */ 
  _listen_handler = io_hook( _listen_fd, IO_READ, _server_accept, NULL, NULL );
}

void server_close        () {
  if( _listen_fd == -1 ) return;
  io_unhook( _listen_handler );
  close( _listen_fd );
  _listen_fd = -1;
}

void server_write        ( Sock_t *   sock,
			   gpointer   data,
			   guint      len ) {
  void (*old)( int );
  gint r, err;

  /* sanity check */
  if( !sock || !data || !len || sock != _socks[sock->fd] )
    return;

  old = signal( SIGPIPE, SIG_IGN );
  r = write( sock->fd, data, len );
  err = errno;
  signal( SIGPIPE, old );

  /* do something with error ? */
}

void server_send         ( Sock_t *   sock,
			   gint       col,
			   gchar *    text ) {
  if( !sock || !text || sock->mute )
    return;
  if( sock->colors && col >= 0 && col < NUM_COLS ) {
    server_write( sock, colors[ col ], strlen( colors[ col ] ) );
    server_write( sock, text, strlen( text ) );
    server_write( sock, colors[ COL_NORMAL ], strlen( colors[ COL_NORMAL ] ) );
  } else
    server_write( sock, text, strlen( text ) );
}

void server_sendf        ( Sock_t *   sock,
			   gint       col,
			   gchar *    format,... ) {  
  gchar * text;
  va_list args;
  va_start( args, format );
  text = g_strdup_vprintf( format, args );
  va_end( args );
  if( text ) {
    server_send( sock, col, text );
    g_free( text );
  }
}

/* call `func' on each connected socket */
void server_foreach      ( SockFunc_t func,
			   gpointer   data ) {
  gint id;
  for( id = 0; id < FD_SETSIZE; id++ ) {
    if( _socks[id] )
      func( _socks[id], data );
  }
}

/* call `func' on each connected socket */
Sock_t * server_find( SockFindFunc_t func,
		      gpointer       data ) {
  gint id;
  for( id = 0; id < FD_SETSIZE; id++ ) {
    if( _socks[id] ) {
      if( func( _socks[id], data ) )
	return _socks[id];
    }
  }
  return NULL;
}

/* send `text' to all connected clients */
void server_broadcast    ( gint       col,
			   gchar *    text ) {
  gint id;
  for( id = 0; id < FD_SETSIZE; id++ ) {
    if( _socks[id] )
      server_send( _socks[id], col, text );
  }
}

void server_broadcastf   ( gint       col,
			   gchar *    format,... ) {
  gchar * text;
  va_list args;
  va_start( args, format );
  text = g_strdup_vprintf( format, args );
  va_end( args );
  if( text ) {
    server_broadcast( col, text );
    g_free( text );
  }
}

/* turn server-controlled echo on/off */
void server_echo         ( Sock_t *   sock,
			   gboolean   echo ) {
  gchar code[4] = { IAC, 0, TELOPT_ECHO, 0 };
  code[1] = ( echo ? WILL : WONT );
  server_write( sock, code, 4 );
}

void server_mute         ( Sock_t *   sock,
			   gboolean   mute ) {
  sock->mute = mute;
}

void server_ping         ( Sock_t * sock ) {
  guchar msg[2] = { IAC, AYT };
  server_write( sock, msg, 2 );
}

/* */
void server_disconnect   ( Sock_t *   sock ) {
  /* sanity check */
  if( !sock || sock != _socks[sock->fd] )
    return;
  _socks[sock->fd] = NULL;
  _num_clients--;
  if( _disconnect_func )
    _disconnect_func( sock, _disconnect_data );
  object_unref( (Object_t *)sock );
}

Sock_t * server_get_sock( gint id ) {
  if( id < 0 || id >= FD_SETSIZE )
    return NULL;
  return _socks[ id ];
}

/* call `func' with a new sock and `data' whenever a client
 * connects.
 */
void     server_set_connect_func   ( SockFunc_t func,
				     gpointer   data ) {
  _connect_func = func;
  _connect_data = data;
}

void     server_set_read_func      ( SockFunc_t func,
				     gpointer   data ) {
  _read_func = func;
  _read_data = data;
}

void     server_set_disconnect_func( SockFunc_t func,
				     gpointer   data ) {
  _disconnect_func = func;
  _disconnect_data = data;
}
