/* Hack'n'Slash, A generic MUD engine.
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL tribe
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* token.c: 
 *   my implementation for reading tokens from file or out of a 
 *   string.  a token is either a word with no delimeters in it
 *   or a string that may or may not have delimeters but is quoted
 *   by a quote pair.  the `returned' token is trimmed of the quote
 *   pair if it was quoted.
 */

#include <string.h>

#include <hacknslash/token.h>

gchar * token_get( gchar *        s,
		   const gchar *  delim,
		   const gchar *  quote,
		   gint *         len,
		   const gchar ** end ) {
  const gchar *quote_char;

  /* scan until no delimeters are found */
  while( 1 ) {
    if( *s == '\0' )
      return NULL;
    if( !strchr( delim, *s ) )
      break;
    s++;
  }

  if( quote != NULL && 
      (quote_char = strchr( quote, *s )) != NULL && 
      ((gint)(quote_char - quote) & 1) == 0 ) {
    /* token is quoted, scan until compliment quote character is found */
    quote_char++;
    s++;
    *len = 0;
    while( 1 ) {
      if( s[ *len ] == '\0' || s[ *len ] == *quote_char )
	break;
      (*len)++;
    }
    if( end )
      *end = &(s[ (*len) + 1 ]);
    
  } else {
    /* token is not quoted */
    *len = 1;
    while( 1 ) {
      if( s[ *len ] == '\0' || strchr( delim, s[ *len ] ) )
	break;
      (*len)++;
    }
    if( end )
      *end = &(s[ (*len) ]);
  }

  return s;
}



gint token_copy_f( FILE *        f,
		   const gchar * delim,
		   const gchar * quote,
		   gchar *       buf,
		   gint          len ) {

  gint ch, i;
  gchar * quote_ch = NULL;

  if( !f )
    return -1;

  /* skip space at beginning */
  while( 1 ) {
    ch = fgetc( f );
    //    fprintf( stdout, "`%c' ", ch );
    if( feof( f ) )
      return -1;
    if( NULL != (quote_ch = strchr( quote, ch ) ) &&
	     !(((gint)(quote_ch - quote)) & 1) ) {
      quote_ch++;
      break;
    }
    if( !strchr( delim, ch ) )
      break;
  }

  if( quote_ch ) {
    /* quoted */
    i = 0;
    while( 1 ) {
      ch = fgetc( f );
      if( feof( f ) ) {
	if( i <= len-1 )
	  buf[i] = 0;
	break;
      }
      if( ch == *quote_ch ) {
	if( i <= len-1 )
	  buf[i] = 0;
	break;
      }
      if( i < len-1 ) {
	buf[i] = ch;
	i++;
	if( i == len-1 )
	  buf[i] = 0;
      }
    }

  } else {
    /* not quoted */
    if( len > 1 )
      buf[0] = ch;
    else
      buf[0] = 0;
    i = 1;
    while( 1 ) {
      ch = fgetc( f );
      if( feof( f ) ) {
	if( i <= len-1 )
	  buf[i] = 0;
	break;
      }
      if( strchr( delim, ch ) ) {
	if( i <= len-1 )
	  buf[i] = 0;
	fseek( f, (long)(-1), SEEK_CUR );
	break;
      }
      if( i < len-1 ) {
	buf[i] = ch;
	i++;
	if( i == len-1 )
	  buf[i] = 0;
      }
    }
  }
  return i;
}

/* sets file pointer for `f' to beginning of next token in `f',
 * returns length of the token on success, -1 on failure.
 */
gint token_get_f( FILE *        f,
		  const gchar * delim,
		  const gchar * quote ) {
  gint ch, len;
  gchar * quote_ch = NULL;

  if( !f )
    return -1;

  /* skip space at beginning */
  while( 1 ) {
    ch = fgetc( f );
    if( feof( f ) )
      return -1;
    if( !strchr( delim, ch ) )
      break;
    else if( NULL != (quote_ch = strchr( quote, ch ) ) &&
	     !((gint)(quote_ch - quote) & 1) ) {
      quote_ch++;
      break;
    }
  }

  if( quote_ch ) {
    /* quoted */
    len = 0;
    while( 1 ) {
      ch = fgetc( f );
      if( feof( f ) )
	break;
      if( ch == *quote_ch )
	break;
      len++;
    }

  } else {
    /* not quoted */
    len = 1;
    while( 1 ) {
      ch = fgetc( f );
      if( feof( f ) || strchr( delim, ch ) ) 
	break;
      len++;
    }
  }
  if( 0 > fseek( f, (long)(-(len+1)), SEEK_CUR ) )
    perror( "fseek" );
  return len;
}

